# coding: utf-8

import time

def f1(x, y):
    return x+y

def f2(x, y):
    return x-y

def f(a, b, c):
    time.sleep((5.6-(3.1*a+1.9*b+.5*c))*0.1)
    if a==0:
        x = 1
    elif a==1:
        x = 7

    if b==0:
        y = x
    elif b==1:
        y = 9

    if c==0:
        z = f1(x,y)
    elif c==1:
        z = f2(x,y)

    return z



# coding: utf-8

from matplotlib import pyplot as plt
import numpy as np
import os
import shutil

from pytest_pitch import khuller_moss_naor as KMN
from pytest_donde.record import Record

''' Demo script to generate (time, coverage) statistics.

This demonstrates the use of pytest_pitch as a library rather than as a pytest plugin.
While pytest is still run in a pre-step, plugin itself is unused.

This can also be used to analyze and compare new algorithmic approaches
without (or before) implementing a new plugin.

Additional dependencies may occur as this is not considered as a functional part of the package.
'''

class KMNBenchmarker:

    def __init__(self, pytest_cmd, src_dir, fname_json):
        self.pytest_cmd = pytest_cmd
        self.src_dir = src_dir
        self.fname_json = fname_json

    def record(self):
        # run donde recorder
        os.system(f'{self.pytest_cmd} --donde={self.src_dir}')
        # backup results
        shutil.copy('donde.json', self.fname_json)

    def visualize(self, record, example_name=None, fname=None,
                  plot_ff=True, plot_kmn=True, plot_normal=True):
        T_flow, C_flow = self.compute_session_history_normal(record)
        Tmax = T_flow[-1]
        Cmax = C_flow[-1]

        if plot_normal:
            self._visualize_conv_plt(T_flow, C_flow, Tmax, Cmax, '-k',
                                     label='pytest')

        if plot_ff:
            T_ff, C_ff = self.evaluate_data_series_fast_first(record)
            self._visualize_conv_plt(T_ff, C_ff, Tmax, Cmax, '-', c='lime',
                                     label='pytest --fast-first')

        if plot_kmn:
            T_kmn, C_kmn = self.compute_session_history_kmn(record, Tmax+0.1)
            self._visualize_conv_plt(T_kmn, C_kmn, Tmax, Cmax, '-', c='lime',
                                     label='pytest --pitch')

        plt.xlabel('time $T/T_{tot}$')
        plt.ylabel('coverage $C/C_{tot}$')

        title = ''
        if example_name:
            title += example_name + '\n'
        title += '$T_{{tot}}={:d}$ sec\t'.format(int(Tmax))
        title += '$C_{{tot}}={}$ LOC'.format(Cmax)

        plt.title(title)
        plt.legend()

        if fname is None:
            plt.show()
        else:
            plt.savefig(fname)

    def compute_session_history_normal(self, record):
        nodeids = record.nodeids()
        return self.compute_session_history(record, nodeids)

    def evaluate_data_series_fast_first(self, record):
        # simulate pytest-fast-first behavior
        nodeids = list(sorted(record.nodeids(),
                              key=record.nodeid_to_duration.__getitem__))
        return self.compute_session_history(record, nodeids)

    def compute_session_history_kmn(self, record, budget):
        nodeids, _, _ = KMN.algorithm(record.nodeid_to_duration, record.nodeid_to_lindices, budget)
        return self.compute_session_history(record, nodeids)

    def compute_session_history(self, record, nodeids):
        # compute (time, coverage) value pairs from (0,0) to (1,1)
        # with increasing values for successively adding another nodeid
        T, C = [], []
        for ind in range(len(nodeids)+1):
            nodeids_sub = nodeids[:ind]
            t, c = self.evaluate_data_point_for_test_selection(record, nodeids_sub)
            T.append(t)
            C.append(c)
        return T, C

    def evaluate_data_point_for_test_selection(self, record, nodeids):
        duration = self._compute_duration(record, nodeids)
        coverage = self._compute_coverage(record, nodeids)
        return (duration, coverage)

    def _visualize_conv_plt(self, T, C, Tmax, Cmax, *a, label='', **kw):
        Tn = np.array(T) / Tmax
        Cn = np.array(C) / Cmax
        for i in range(len(T)):
            if Cn[i] > 0.99:
                pos = i
                break
        else:
            pos = len(T)-1

        label += '\n$0.99 C_{{tot}}$ @ ${:02.2f}T_{{tot}}$'.format(Tn[pos])
        plt.plot(Tn, Cn, *a, **kw, label=label)
        return Tn, Cn, Tn[pos]

    def _compute_duration(self, record, nodeids):
        return sum(record.nodeid_to_duration[nid] for nid in nodeids)

    def _compute_coverage(self, record, nodeids):
        covered = set()
        for nodeid in nodeids:
            covered.update(record.nodeid_to_lindices[nodeid])
        return len(covered)

if __name__ == '__main__':
    import sys

    pytest_cmd, src_dir, fname_json = sys.argv[1:]
    b = KMNBenchmarker(pytest_cmd, src_dir, fname_json)

    # you may want to comment out this line to reuse the same donde.json,
    # if you just want to change the processing part
    b.record()

    record = Record.from_file(fname_json)
    b.visualize(record, example_name='example project', plot_ff=0)



# -*- coding: utf-8 -*-

import pytest

import os
import shutil

ROOT_PATH = os.path.dirname(__file__)+'/..'

def _setup_example(testdir, example_dir):
    shutil.copytree(os.path.join(ROOT_PATH, example_dir, 'src'),
                    os.path.join(str(testdir), 'src'))
    shutil.copyfile(os.path.join(ROOT_PATH, example_dir, 'test_src.py'),
                    os.path.join(str(testdir), 'test_src.py'))


@pytest.mark.parametrize('example_dir, pytest_args_donde, pytest_args_pitch, expected_nodeids_ordered, expected_nodeids_random', [
    ('examples/example01', tuple(), tuple(), [
            'test_src.py::test_f[1-1-1--2]',
            'test_src.py::test_f[1-1-0-16]',
            'test_src.py::test_f[1-0-1-0]',
            'test_src.py::test_f[0-1-1--8]',

    ], [
            'test_src.py::test_f[1-0-0-14]',
            'test_src.py::test_f[0-0-0-2]',
            'test_src.py::test_f[0-0-1-0]',
            'test_src.py::test_f[0-1-0-10]',
    ]),
    ('examples/example01', tuple(), ('-k 1-1-1', ), [
            'test_src.py::test_f[1-1-1--2]',

    ], [
    ]),
    ('examples/example01', ('-k 1-1-1', ), tuple(), [
            'test_src.py::test_f[1-1-1--2]',

    ], [
    ]),
    ('examples/example01', ('-k 1-1-1', ), ('-k 1-1-1', ), [
            'test_src.py::test_f[1-1-1--2]',

    ], [
    ]),
    ('examples/example01', tuple(), ('-k 1-1', ), [
            'test_src.py::test_f[1-1-1--2]',
            'test_src.py::test_f[1-1-0-16]',
            'test_src.py::test_f[0-1-1--8]',
    ], [
    ]),
    ('examples/example01', tuple(), ('-k 1-0', ), [
            'test_src.py::test_f[1-1-0-16]',
            'test_src.py::test_f[1-0-1-0]',
    ], [
            'test_src.py::test_f[1-0-0-14]',
            'test_src.py::test_f[0-1-0-10]',
    ]),
])
def test_recorder_example01(testdir, example_dir, pytest_args_donde, pytest_args_pitch, expected_nodeids_ordered, expected_nodeids_random):

    _setup_example(testdir, example_dir)

    testdir.runpytest('-v', '--donde=src', *pytest_args_donde)
    result = testdir.runpytest('-v', '--pitch', *pytest_args_pitch)

    # FIXME find out why expressions won't work with closing bracket ...
    lines_ordered = ['*{}*PASSED*'.format(nodeid.rstrip(']')) for nodeid in expected_nodeids_ordered]
    lines_random = ['*{}*PASSED*'.format(nodeid.rstrip(']')) for nodeid in expected_nodeids_random]

    # they must appear exactly in that order
    result.stdout.fnmatch_lines(lines_ordered, consecutive=True)

    # these remain with zero coverage gain, so their order is random,
    # but we know that individually, each of them must occur after the ordered lines
    for line in lines_random:
        result.stdout.fnmatch_lines(lines_ordered + [line], consecutive=False)



# coding: utf-8

def algorithm(test_id_to_duration, test_id_to_hit_indices, budget):
    # see p. 3, Algorithm 1 of https://doi.org/10.1016/S0020-0190(99)00031-9
    # for reference, we use rather cryptic math notation here

    i_to_Ti = [T for _, T in sorted(test_id_to_hit_indices.items())]
    i_to_ci = [t for _, t in sorted(test_id_to_duration.items())]
    index_to_test_id = {index: test_id for index, test_id in enumerate(sorted(test_id_to_duration))}
    L = budget

    # begin
    G = set()
    G_indices = []
    U_indices = set(range(len(test_id_to_duration)))
    C = 0
    COV = 0
    while U_indices:
        best_i = None
        best_ci = None
        best_Wipci = -1
        best_cov_gain = None
        for i in sorted(U_indices): # ensures determinism, wont be worse than log(size)
            Wip = len(i_to_Ti[i].difference(G))
            ci = i_to_ci[i]
            Wipci = Wip / ci
            if Wipci > best_Wipci:
                best_i = i
                best_Wipci = Wipci
                best_ci = ci
                best_cov_gain = Wip
        if C + best_ci <= L:
            C += best_ci
            G_indices.append(best_i)
            G.update(i_to_Ti[best_i])
            COV += best_cov_gain
        U_indices.remove(best_i)

    test_ids_ordered = [index_to_test_id[index] for index in G_indices]
    return test_ids_ordered, C, COV



# -*- coding: utf-8 -*-

from pytest_donde.record import Record

from . import khuller_moss_naor as KMN

def pytest_addoption(parser):
    group = parser.getgroup('pitch')
    group.addoption(
        '--pitch',
        action='store_true',
        dest='pitch_active',
        default=None,
        help='reorder tests for fast increase of coverage using donde session record.',
    )
    group.addoption(
        '--pitch-in',
        action='store',
        dest='pitch_donde_json_path',
        metavar='PATH',
        default='donde.json',
        help='set custom PATH to donde record file, default is "donde.json".',
    )

def pytest_configure(config):
    if config.getoption('pitch_active'):
        path = config.getoption('pitch_donde_json_path')
        config.pluginmanager.register(PitchSelectorPlugin(path))

class PitchSelectorPlugin:

    def __init__(self, path_input):
        self._nodeids_changed = False
        self._nodeid_order = self._compute_optimal_order(path_input)

    def _compute_optimal_order(self, path_input):
        record = Record.from_file(path_input)

        budget = sum(record.nodeid_to_duration.values())
        budget += 0.1 # circumvent float precision issues to ensure we catch all tests

        print('[pitch] computing optimal test order ...')
        nodeids, _, _ = KMN.algorithm(record.nodeid_to_duration, record.nodeid_to_lindices, budget)
        print('[pitch] computing optimal test order ... done')
        return nodeids

    def pytest_collection_modifyitems(self, items, config):

        def key(item):
            try:
                return self._nodeid_order.index(item.nodeid)
            except ValueError:
                if not self._nodeids_changed:
                    self._nodeids_changed = True
                    # FIXME learn about pytest warning mechanism
                    path = config.getoption('pitch_donde_json_path')
                    print(f'[pitch] the test item {item.nodeid} was not registered in {path} '
                          'and will be placed at the start of the session. '
                          'Further events of this type will not be logged. '
                          'If You did not expect this, consider refreshing your donde record '
                          'to match your current test session.')
                return -1

        items[:] = sorted(items, key=key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import codecs
from setuptools import setup, find_packages


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()

def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

example_files = []
example_files += package_files('examples')
example_files += package_files('scripts')

from setuptools import find_packages
setup(
    name='pytest_pitch',
    version='1.0.5',
    author='Clemens Löbner',
    author_email='mikamove@posteo.de',
    maintainer='Clemens Löbner',
    maintainer_email='mikamove@posteo.de',
    license='MIT',
    url='https://github.com/mikamove/pytest-pitch',
    description='runs tests in an order such that coverage increases as fast as possible',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    py_modules=['pytest_pitch'],
    python_requires='>=3.8',
    install_requires=[
        'pytest>=7.3.1',
        'pytest-donde>=1.1.0',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    entry_points={
        'pytest11': [
            'pitch = pytest_pitch.plugin',
        ],
    },
    packages=find_packages(where='.'),
    package_dir={'': '.'},
    package_data={'': example_files},
 )
